# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test library for picatrix."""
from typing import Any
from typing import Text

from IPython.terminal.interactiveshell import TerminalInteractiveShell
from IPython.testing.globalipapp import start_ipython

from picatrix.lib import manager


manager.MagicManager.is_test = True


class InteractiveTest:
  """Class that helps with interactive tests for picatrix."""

  _shell: TerminalInteractiveShell = None

  @classmethod
  def _init_shell(cls):
    """Initialize the shell."""
    ip = start_ipython()

    ip.run_cell(raw_cell='from picatrix import notebook_init')
    ip.run_cell(raw_cell='notebook_init.init()')
    cls._shell = ip

  @classmethod
  def get_shell(cls) -> TerminalInteractiveShell:
    """Return an interactive shell, initializing it for the first time."""
    if cls._shell:
      return cls._shell
    cls._init_shell()
    return cls._shell

  @classmethod
  def run_and_compare(cls, code: Text, expected_return: Any):
    """Run code in shell, get returns and assert against expected returns."""
    shell = cls.get_shell()
    res = shell.run_cell(raw_cell=code)
    assert res.success
    assert res.result == expected_return



"""Timesketch magic module.

This module conatins an implementation for the Picatrix Timesketch integration.
It enables colab/jupyter to send and receive data from a Timesketch sketch.
"""
import datetime
import logging
import os

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Text
from typing import Union

import dateutil.parser

from dfdatetime import time_elements

import pandas as pd

from timesketch_api_client import aggregation as api_aggregation
from timesketch_api_client import config
from timesketch_api_client import view as api_view
from timesketch_api_client import sketch as api_sketch
from timesketch_api_client import story as api_story
from timesketch_api_client import timeline as api_timeline
from timesketch_import_client import helper
from timesketch_import_client import importer

from picatrix.lib import framework
from picatrix.lib import state
from picatrix.lib import utils


def _add_date_chip(
    query_filter: Dict[str, Any], start_time: Text, end_time: Text):
  """Adds a date range chip into a query filter.

  Args:
    query_filter (dict): the query filter.
    start_time (str): a date string with the start of the filter.
    end_time (str): a date string with the end of the filter.
  """
  date_string = f'{start_time},{end_time}'
  query_filter.setdefault('chips', [])
  query_filter['chips'].append({
      'field': '',
      'type': 'datetime_range',
      'value': date_string})


def _fix_return_fields(
    return_fields: Union[Text, List[Text], None]) -> Optional[Text]:
  """Returns a fixed string of return fields.

  The Timesketch API expects the return fields to be a comma separated
  string with no spaces in it and it cannot start with a quotation mark.

  Args:
    return_fields (str|list): a list of strings or a string with the
        return fields.

  Returns:
    A string that can be passed to the Timesketch API.
  """
  if isinstance(return_fields, str):
    return_fields = return_fields.split(',')

  if return_fields is None:
    return None

  return_field_list = []
  for field in return_fields:
    field = field.replace('\'', '')
    field = field.replace('"', '')
    field = field.strip()

    return_field_list.append(field)

  return ','.join(return_field_list)


def _label_search(
    label: Text,
    return_fields: Optional[Text] = '',
    max_entries: Optional[int] = None) -> pd.DataFrame:
  """Returns a DataFrame with all events that have a certain label.

  Args:
    label (str): a string representing the label to search for.
    return_fields (str): a comma separated string with all the fields that
        should be returned. If not provided all fields will be returned.
    max_entries (int): an integer with the maximum number of entries to return.

  Returns:
      A pandas DataFrame with the search results.
  """
  connect()
  state_obj = state.state()

  sketch = state_obj.get_from_cache('timesketch_sketch')
  if not sketch:
    print('No data, not connected to a sketch.')
    return pd.DataFrame()

  if not return_fields:
    return_fields = '*'

  return sketch.search_by_label(
      label,
      return_fields=_fix_return_fields(return_fields),
      max_entries=max_entries,
      as_pandas=True)


def connect(
    ignore_sketch: bool = False,
    force_switch: bool = False):
  """Check if Timesketch has been set up and connect if it hasn't.

  Args:
    ignore_sketch (optional): if set to True sketch check is ignored.
    force_switch (optional): if set to True then a new client will be created,
        irrelevant if another client was stored.

  Raises:
    ValueError: if Timesketch is not properly configured.
  """
  state_obj = state.state()

  client = state_obj.get_from_cache('timesketch_client')
  sketch = state_obj.get_from_cache('timesketch_sketch')

  if client and force_switch:
    state_obj.remove_from_cache('timesketch_client')
    client = None

  if client:
    if ignore_sketch:
      return

    if sketch:
      return

    raise ValueError(
        'No sketch configured, either create a new one using '
        '%timesketch_create_sketch or assign an already existing '
        'one using %timesketch_set_active_sketch <sketch_id>')

  client = config.get_client()
  if not client:
    raise ValueError('Unable to connect to Timesketch')

  state_obj.add_to_cache('timesketch_client', client)


def get_context_date(
    date_string: Text, minutes: Optional[int] = 0, seconds: Optional[int] = 90,
    return_fields: Text = '') -> pd.DataFrame:
  """Return time context surrounding a single data frame row.

  Args:
    date_string (str): a date string to gather context around, in the form
        of %Y-%m-%dT%H:%M:%S%z.
    minutes (int): if provided used as number of minutes before
        and after the provided date string to provide context for. If minutes
        are provided seconds argument is ignored.
    seconds (int): if minutes is not provided seconds is used to determine
        the number of seconds before and after the date string to provide
        context for. Default value is 90 seconds.
    return_fields (str): string with comma separated names of fields to
        return back.

  Returns:
    pd.DataFrame: returns a data frame with all events in Timesketch
        that occurred within the timeframe supplied to the function.
  """
  if minutes:
    seconds = 60 * minutes

  date_object = dateutil.parser.parse(date_string)
  start_date = date_object - datetime.timedelta(seconds=seconds)
  end_date = date_object + datetime.timedelta(seconds=seconds)

  start_string = start_date.strftime('%Y-%m-%dT%H:%M:%S%z')
  end_string = end_date.strftime('%Y-%m-%dT%H:%M:%S%z')

  query_filter = {
      'time_start': None,
      'time_end': None,
      'indices': ['_all'],
      'order': 'asc',
  }
  _add_date_chip(query_filter, start_string, end_string)
  return_fields = _fix_return_fields(return_fields)
  return query_timesketch(
      '*',
      return_fields=return_fields,
      query_filter=query_filter)


def get_context_row(
    row: pd.Series, minutes: Optional[int] = 0,
    seconds: Optional[int] = 90,
    return_fields: str = '') -> pd.DataFrame:
  """Return time context surrounding a single data frame row.

  Args:
    row (pandas.core.series.Series): a single row to use to extract the
        date from.
    minutes (int): if provided used as number of minutes before
        and after the provided date string to provide context for. If minutes
        are provided seconds argument is ignored.
    seconds (int): if minutes is not provided seconds is used to determine
        the number of seconds before and after the date string to provide
        context for. Default value is 90 seconds.
    return_fields (string): string with comma separated names of fields to
        return back.

  Returns:
    pandas.DataFrame: returns a data frame with all events in Timesketch
        that occurred within the timeframe supplied to the function.
  """
  if not isinstance(row, pd.Series):
    return pd.DataFrame()

  if 'datetime' not in row:
    return pd.DataFrame()

  date_string = row.datetime

  if hasattr(date_string, 'year') and hasattr(date_string, 'strftime'):
    # We have a pandas date object.
    date_string = date_string.strftime('%Y-%m-%dT%H:%M:%S%z')

  return get_context_date(
      date_string, minutes=minutes, seconds=seconds,
      return_fields=return_fields)


def is_view_object(input_object: Any) -> bool:
  """Returns a boolean whether an object is a view object or not.

  Args:
    input_object (object): an object that is to be tested whether or
        not it is a View object.

  Returns:
    True if it is a View object, False otherwise.
  """
  if not hasattr(input_object, 'lazyload_data'):
    return False

  if not hasattr(input_object, 'query_dsl'):
    return False

  if not hasattr(input_object, 'name'):
    return False

  if not hasattr(input_object, 'query_filter'):
    return False

  return True


def format_data_frame_row(
    row: pd.Series, format_message_string: Text) -> pd.DataFrame:
  """Return a formatted data frame using a format string."""
  return format_message_string.format(**row)


def get_sketch_details(sketch_id: Optional[int] = 0) -> Text:
  """Return back details about an existing sketch.

  Args:
    sketch_id (int): the sketch ID to check.

  Returns:
    string containing information about the sketch.
  """
  sketch = None
  state_obj = state.state()

  if not sketch_id:
    sketch = state_obj.get_from_cache('timesketch_sketch')
    if not sketch:
      return 'Need to provide a sketch id, no sketch provided.'

  connect(ignore_sketch=True)
  client = state_obj.get_from_cache('timesketch_client')
  if not sketch:
    sketch = client.get_sketch(sketch_id)

  try:
    _ = sketch.name
  except KeyError:
    return 'TS server returned no information back about the sketch'

  return_string_list = []
  return_string_list.append(f'Name: {sketch.name}')
  return_string_list.append(f'Description: {sketch.description}')
  return_string_list.append(f'ID: {sketch.id}')
  return_string_list.append('')
  return_string_list.append('Active Timelines:')

  for timeline in sketch.list_timelines():
    objects = timeline.data.get('objects', [])
    if not objects:
      continue

    data = objects[0]
    description = data.get('description', 'No description')
    created_at = data.get('created_at', 'N/A')
    return_string_list.append(
        f'{timeline.name} [{description}] -> {created_at}')

  return '\n'.join(return_string_list)


def set_active_sketch(sketch_id: int):
  """Set the active sketch."""
  connect(ignore_sketch=True)
  state_obj = state.state()

  client = state_obj.get_from_cache('timesketch_client')

  sketch = client.get_sketch(sketch_id)
  state_obj.add_to_cache('timesketch_sketch', sketch)


# pylint: disable=unused-argument
@framework.picatrix_magic
def timesketch_get_sketch(data='') -> api_sketch.Sketch:
  """Returns the Timesketch sketch object.

  Args:
    data (str): Not used, should be empty but is ignored.

  Returns:
    The selected sketch object (instance of api_sketch.Sketch).
  """
  connect()
  state_obj = state.state()
  return state_obj.get_from_cache('timesketch_sketch')


@framework.picatrix_magic
def timesketch_add_manual_event(
    data: Text, timestamp: Optional[int] = 0,
    date_string: Optional[Text] = '',
    timestamp_desc: Optional[Text] = '',
    attributes: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None) -> Dict[str, str]:
  """Add a manually generated event to the sketch.

  Args:
    data (str): The message string for for the event to be generated.
    timestamp (int): Optional timestamp in either seconds since Epoch or
        microseconds since Epoch.
    date_string (str): An optional date time as a human readable string. If
        neither date_string nor timestamp is provided then the current timestamp
        will be used as the time of the event.
    timestamp_desc (str): Optional timestamp description field.
    attributes (dict): Optional dict which contains extra attributes to add
        to the manual event.
    tags (list): Optional list of tags to add to the manual event.

  Returns:
    Dictionary with query results.
  """
  connect()
  state_obj = state.state()
  sketch = state_obj.get_from_cache('timesketch_sketch')
  if not sketch:
    print('Not able to connect to a sketch.')
    return {}

  # Default timestamp.
  date_obj = datetime.datetime.now(datetime.timezone.utc)
  date = date_obj.isoformat()
  if timestamp:
    try:
      date_obj = datetime.datetime.fromtimestamp(
          timestamp, datetime.timezone.utc)
    except ValueError:
      date_obj = datetime.datetime.fromtimestamp(
          timestamp / 1e6, datetime.timezone.utc)
    date = date_obj.isoformat()
  elif date_string:
    elements = time_elements.TimeElements()
    if 'T' in date_string:
      try:
        elements.CopyFromStringISO8601(date_string)
      except ValueError:
        logging.error(
            'Unable to convert date string, is it really in ISO 8601 format?')
        return {}
    try:
      elements.CopyFromString(date_string)
    except ValueError:
      try:
        elements.CopyFromStringRFC1123(date_string)
      except ValueError:
        logging.error(
            'Unable to convert date string, needs to be in ISO 8601, 1123 or '
            'in the format YYYY-MM-DD hh:mm:ss.######[+-]##:##')
        return {}
    date = elements.CopyToDateTimeStringISO8601()

  if not timestamp_desc:
    timestamp_desc = 'Event Logged'

  if not isinstance(tags, (tuple, list)):
    tags = []

  if not isinstance(attributes, dict):
    attributes = {}

  if not date:
    logging.error('Unable to convert date string, please check it.')
    return {}

  return sketch.add_event(
      data, date, timestamp_desc, attributes=attributes, tags=tags)


@framework.picatrix_magic
def timesketch_upload_file(data: Text, name: Optional[Text] = ''):
  """Upload a file to Timesketch.

  Args:
    data (str): Path to the file that's about to be uploaded.
    name (str): Name of the timeline.
  """
  if not os.path.isfile(data):
    print('File [{0:s}] does not exist.'.format(data))
    return

  connect()
  state_obj = state.state()
  sketch = state_obj.get_from_cache('timesketch_sketch')
  result = None

  if not name:
    name = data.split(os.sep)[-1]
    first, _, last = name.rpartition('.')
    name = first or last
    name = name.replace(' ', '_').replace('-', '_')

  with importer.ImportStreamer() as streamer:
    streamer.set_sketch(sketch)
    streamer.set_timeline_name(name)

    # Set the file size to 20Mb before the file is split.
    streamer.set_filesize_threshold(20971520)
    streamer.add_file(data)
    result = streamer.response

  if not result:
    print('Unable to upload data.')
    return

  for timesketch_object in result.get('objects', []):
    if not timesketch_object:
      continue
    print('Timeline: {0:s}\nStatus: {1:s}'.format(
        timesketch_object.get('description'),
        ','.join([x.get('status') for x in timesketch_object.get('status')])))


def query_timesketch(
    query: Optional[Text] = None,
    query_dsl: Optional[Text] = None,
    query_filter: Optional[Dict[Text, Any]] = None,
    view: Optional[api_view.View] = None,
    return_fields: Optional[Text] = None,
    start_date: Optional[Text] = '',
    end_date: Optional[Text] = '',
    max_entries: Optional[int] = None,
    indices: Optional[List[Text]] = None) -> pd.DataFrame:
  """Return back a data frame from a Timesketch query.

  Args:
    query (str): the query string to send to Timesketch.
    query_dsl (str): the query DSL to send to Timesketch.
    view (api_view.View): View object.
    return_fields (str): string with comma separated names of fields to
        return back.
    start_date (str): a timestamp in the form of
        'YYYY-MM-DDTHH:MM:SS+00:00' of when to start the search query.
    end_date (str): a timestamp in the form of
        'YYYY-MM-DDTHH:MM:SS+00:00' of when to end the search query.
    max_entries (int): Optional integer denoting a best effort to limit
        the output size to the number of events. Events are read in,
        10k at a time so there may be more events in the answer back
        than this number denotes, this is a best effort. This value defaults
        to 40k events. Setting max_entries to zero will return all entries
        back (as in no limit).
    indices (list): a list of indices to query, if not provided _all
        will be used.

  Returns:
    A data frame with the results gathered from the query.

  Raises:
    KeyError: if the query is sent in without a query, query_dsl or a view.
  """
  connect()
  state_obj = state.state()
  sketch = state_obj.get_from_cache('timesketch_sketch')

  if all([x is None for x in [query, query_dsl, view]]):
    raise KeyError('Need to provide a query, query_dsl or a view')

  if not query_filter:
    query_filter = {
        'time_start': None,
        'time_end': None,
        'order': 'asc'
    }

  if indices:
    query_filter['indices'] = indices
  else:
    query_filter['indices'] = '_all'

  if start_date or end_date:
    _add_date_chip(query_filter, start_date, end_date)

  # If view is being sent in, view needs to be the only parameter to the search.
  if view is not None:
    query = None
    query_dsl = None

  return_fields = _fix_return_fields(return_fields)

  return sketch.explore(
      query_string=query, query_dsl=query_dsl,
      query_filter=query_filter, view=view, return_fields=return_fields,
      max_entries=max_entries, as_pandas=True)


# pylint: disable=unused-argument
@framework.picatrix_magic
def timesketch_list_views(
    data: Optional[Text] = '') -> Dict[str, api_view.View]:
  """List up all available views.

  Args:
    data (str): Not used.

  Returns:
    A dict with a list of available views.
  """
  connect()
  state_obj = state.state()
  sketch = state_obj.get_from_cache('timesketch_sketch')

  return_dict = {}
  for view in sketch.list_views():
    return_dict['{0:d}:{1:s}'.format(view.id, view.name)] = view
  return return_dict


@framework.picatrix_magic
def timesketch_query(
    data: Text,
    fields: Optional[Text] = None,
    timelines: Optional[Text] = None,
    view: Optional[api_view.View] = None,
    start_date: Optional[Text] = '',
    end_date: Optional[Text] = '',
    query_filter: Optional[Dict[Text, Any]] = None,
    max_entries: Optional[int] = 40000) -> pd.DataFrame:
  """Run a Timesketch query using a magic.

  Args:
    data (str): the Timesketch query.
    fields (str): comma separated list of fields to include in the returned
        data frame.
    timelines (str): comma separated list of the names of the timelines.
    view (View): the View object.
    start_date (str): start date of the result set in the form of
        YYYY-MM-DDTHH:MM:SS+00:00.
    end_date (str): end date of the result set in the form of
        YYYY-MM-DDTHH:MM:SS+00:00.
    query_filter (dict): The query filter dict to use.
    max_entries (int): maximum entries of returned object, defaults to 40k.

  Raises:
    ValueError: If a view object is passed in and it's not a proper view
        object.

  Returns:
    DataFrame containing the results of the query.
  """
  connect()
  state_obj = state.state()

  if timelines:
    sketch = state_obj.get_from_cache('timesketch_sketch')
    timeline_list = sketch.list_timelines()
    names = timelines.split(',')
    indices = [
        x.index for x in timeline_list if x.name in names]
  else:
    indices = None

  if view and not is_view_object(view):
    raise ValueError('View is not a view instance, but rather [{0:s}]'.format(
        type(view)))

  return query_timesketch(
      query=data, view=view, return_fields=fields, query_filter=query_filter,
      indices=indices, start_date=start_date, end_date=end_date,
      max_entries=max_entries)


@framework.picatrix_magic
def timesketch_context_date(
    data: Text, minutes: int, seconds: Optional[int] = 90,
    fields: Optional[Text] = '') -> pd.DataFrame:
  """Run a Timesketch context query using a magic.

  This is a magic to run a context query in Timesketch around a
  single date and save them to a dataframe.

  If minutes are provided the seconds will be ignored.

  Args:
    data (str): the date string, in the form of %Y-%m-%dT%H:%M:%S%z.
    minutes (int): number of minutes to include in context.
    seconds (int): number of seconds to include in context, defaults
        to 90 seconds.
    fields (str): optional list of fields to include in the returned data.

  Returns:
    DataFrame containing the all surrounding events.
  """
  connect()
  date = data.strip()

  if minutes:
    seconds = minutes * 60

  return get_context_date(
      date, minutes=minutes, seconds=seconds, return_fields=fields)


@framework.picatrix_magic
def timesketch_context_row(
    data: pd.Series, minutes: int, seconds: Optional[int] = 90) -> pd.DataFrame:
  """Run a Timesketch context query using a magic.

  This is a magic to run a context query in Timesketch around a
  single date and save them to a dataframe.

  If minutes are provided the seconds will be ignored.

  Args:
    data (pandas.core.series.Series): pandas Series that contain the date
        string, in the form of %Y-%m-%dT%H:%M:%S%z.
    minutes (int): number of minutes to include in context.
    seconds (int): number of seconds to include in context, defaults
        to 90 seconds.

  Raises:
    ValueError: if the data field is not of the correct type.

  Returns:
    DataFrame containing the all surrounding events.
  """
  connect()

  if not isinstance(data, pd.Series):
    raise ValueError((
        'This magic expects a pandas Series object, use curly braces '
        '{{var_name}} to expand variable names.'))

  if minutes:
    seconds = minutes * 60

  return get_context_row(data, minutes=minutes, seconds=seconds)


@framework.picatrix_magic
def timesketch_get_timelines(data: Text) -> Dict[str, api_timeline.Timeline]:
  """Magic to get all available timelines for the sketch.

  Args:
    data (str): sketch ID to check for timelines, otherwise the active sketch
        is used.

  Returns:
    A dict with all timelines attached to the sketch, with keys as timeline name
    and value the timeline object (instance of Timeline).
  """
  connect()
  state_obj = state.state()

  if data:
    sketch_id = data.strip()
    client = state_obj.get_from_cache('timesketch_client')
    sketch = client.get_sketch(sketch_id)
  else:
    sketch = state_obj.get_from_cache('timesketch_sketch')

  if not sketch:
    return []

  return {x.name: x for x in sketch.list_timelines()}


@framework.picatrix_magic
def timesketch_create_sketch(
    data: Text, description: Text, set_active: Optional[bool] = False) -> Text:
  """Magic to create a new sketch in Timesketch.

  Args:
    data (str): name of the new sketch.
    description (str): sketch description.
    set_active (bool): whether to set the newly created sketch as the active
        one. Defaults to False.

  Returns:
    str: response string with sketch IDs.
  """
  connect(ignore_sketch=True)
  state_obj = state.state()
  client = state_obj.get_from_cache('timesketch_client')

  name = data.strip()
  sketch = client.create_sketch(name, description)

  if not sketch:
    return 'No response, sketch not created.'

  try:
    _ = sketch.name
  except KeyError:
    return 'It seems like the sketch was not created, verify on TS server.'

  if set_active:
    set_active_sketch(sketch.id)

  return_string_list = []
  return_string_list.append(f'Sketch: {sketch.id}')
  return_string_list.append(f'Name: {sketch.name}')

  data_objects = sketch.data.get('objects')
  if data_objects:
    data = data_objects[0]
    status_dict = data.get('status', [{}])[0]
    creation_time = status_dict.get('created_at', 'N/A')
    return_string_list.append('Creation Time: {0:s}'.format(creation_time))
    status = status_dict.get('status', 'N/A')
    return_string_list.append('Status: {0:s}'.format(status))

  return '\n'.join(return_string_list)


@framework.picatrix_magic
def timesketch_get_views(data: Optional[Text] = '') -> Dict[str, api_view.View]:
  """Get all views from an active sketch.

  Args:
    data (str): the sketch ID used to fetch views from. Defaults to the
        active sketch.

  Returns:
      Dict that contains View objects (instance of View) as value and the
      name of the view as the key.

  Raises:
    ValueError: if Timesketch is not properly configured.
  """
  connect()
  state_obj = state.state()

  if data:
    sketch_id = data.strip()
    if not sketch_id.isdigit():
      raise ValueError('Sketch ID is not a digit.')
    sketch_id = int(sketch_id, 10)
    client = state_obj.get_from_cache('timesketch_client')
    sketch = client.get_sketch(sketch_id)
  else:
    sketch = state_obj.get_from_cache('timesketch_sketch')

  if not sketch:
    raise ValueError('No sketch ID provided.')

  return {x.name: x for x in sketch.list_views()}


@framework.picatrix_magic
def timesketch_query_view(data: api_view.View) -> pd.DataFrame:
  """Makes a Timesketch query using the active sketch and a view.

  Args:
    data (View): the View object to query.

  Returns:
    DataFrame containing the results of the query in line mode, otherwise None.

  Raises:
    ValueError: if View object is not presented.
  """
  connect()
  if not data:
    raise ValueError('No View object passed in.')

  if not is_view_object(data):
    raise ValueError((
        'Need to pass in a View object, remember to use curly braces '
        'surrounding variable names.'))

  return query_timesketch(view=data, max_entries=0)


@framework.picatrix_magic
def timesketch_set_active_sketch(data: Text):
  """Sets the active sketch.

  Args:
    data (str): the sketch ID to set configure Timesketch for.
  """
  sketch_id = int(data.strip(), 10)
  set_active_sketch(sketch_id)
  utils.clear_notebook_output()


@framework.picatrix_magic
def timesketch_upload_data(
    data: pd.DataFrame,
    name: Optional[Text] = '',
    format_message_string: Optional[Text] = ''):
  """Upload a data frame to TimeSketch.

  Args:
    data (pandas.core.frame.DataFrame): the DataFrame to upload.
    name (str): the name used for the timeline in Timesketch.
    format_message_string (str): formatting string for the message column of
        the data frame, eg: "{src_ip:s} to {dst_ip:s}, {bytes:d} bytes
        transferred"'

  Raises:
    ValueError: if the dataframe cannot be uploaded to Timesketch or the
        data is invalid.
  """
  if not isinstance(data, pd.DataFrame):
    raise ValueError((
        'The data attribute is not a pandas DataFrame, please use curly '
        'braces to expand variables.'))

  if not name:
    name = 'unknown_timeline'

  connect()
  state_obj = state.state()

  sketch = state_obj.get_from_cache('timesketch_sketch')
  if not sketch:
    raise ValueError('Unable to upload data frame, need to set sketch.')

  result = None

  import_helper = helper.ImportHelper()
  with importer.ImportStreamer() as streamer:
    streamer.set_sketch(sketch)

    if 'data_type' not in data:
      data_type = utils.ask_question(
          'What is the value of [data_type]?', input_type=str)
      if data_type:
        streamer.set_data_type(data_type)
    else:
      data_types = data.data_type.unique()
      data_type = data_types[0]

    columns = list(data.columns)
    streamer.set_config_helper(import_helper)
    import_helper.configure_streamer(
        streamer, data_type=data_type, columns=columns)

    if format_message_string:
      streamer.set_message_format_string(format_message_string)

    streamer.set_timeline_name(name)

    streamer.add_data_frame(data)
    result = streamer.response

  if not result:
    print('Unable to upload data.')
    return

  for timesketch_object in result.get('objects', []):
    if not timesketch_object:
      continue
    print('Timeline: {0:s}\nStatus: {1:s}'.format(
        timesketch_object.get('description'),
        ','.join([x.get('status') for x in timesketch_object.get('status')])))


# pylint: disable=unused-argument
@framework.picatrix_magic
def timesketch_list_timelines(data: Optional[Text] = '') -> Text:
  """Returns a string with information about timelines and analyzer results."""
  connect()
  state_obj = state.state()
  sketch = state_obj.get_from_cache('timesketch_sketch')
  if not sketch:
    return 'No data, not connected to a sketch.'

  timelines = sketch.list_timelines()
  return_lines = []

  for timeline in timelines:
    return_lines.append('         --- Timeline: {0:^s} ---'.format(
        timeline.name))
    return_lines.append('-'*80)
    return_lines.append('')

    data = timeline.data
    for index, data_object in enumerate(data.get('objects', [])):
      return_lines.append('Object nr: {0:d}'.format(index + 1))

      description = data_object.get('description')
      if description:
        return_lines.append('Description: {0:s}'.format(description))
        return_lines.append('')

      status_list = data_object.get('status', [])
      status = 'unknown'
      if status_list:
        status_dict = status_list[0]
        status = status_dict.get('status', 'N/A')
        return_lines.append('Status: {0:s}'.format(status))
        return_lines.append('')

      analyzer_output = data_object.get('searchindex', {}).get('description')
      if status == 'ready':
        return_lines.append('Analyzer Reports:')
        analyzer_lines = []
        for x in analyzer_output.split('\n'):
          if x and x not in analyzer_lines:
            analyzer_lines.append(x)
        for line_number, line in enumerate(analyzer_lines):
          return_lines.append(
              '  [{0:d}] {1:s}'.format(line_number + 1, line.strip()))
      else:
        return_lines.append('Error Message:')
        return_lines.append(analyzer_output)

    return_lines.append('='*80)
    return_lines.append('')
  return '\n'.join(return_lines)


@framework.picatrix_magic
def timesketch_get_aggregations(
    data: Optional[Text] = '') -> Dict[Text, api_aggregation.Aggregation]:
  """Get all aggregations from an active sketch.

  Args:
    data (str): the sketch ID used to fetch aggregations from. Defaults to the
        active sketch.

  Returns:
      Dict that contains Aggregation objects (instance of Aggregation) as
      value and the name of the aggregation as the key.

  Raises:
    ValueError: if Timesketch is not properly configured.
  """
  connect()
  state_obj = state.state()

  if data:
    sketch_id = data.strip()
    if not sketch_id.isdigit():
      raise ValueError('Sketch ID is not a digit.')
    sketch_id = int(sketch_id, 10)
    client = state_obj.get_from_cache('timesketch_client')
    sketch = client.get_sketch(sketch_id)
  else:
    sketch = state_obj.get_from_cache('timesketch_sketch')

  if not sketch:
    raise ValueError('No sketch ID provided.')

  aggregations = sketch.list_aggregations()
  return {x.name: x for x in aggregations}


# pylint:disable=unused-argument
@framework.picatrix_magic
def timesketch_available_aggregators(
    data: Optional[Text] = '') -> pd.DataFrame:
  """Returns a data frame with information about available aggregators.

  Args:
    data (str): not used.

  Returns:
    A pandas DataFrame with information about available aggregators.
  """
  connect()
  state_obj = state.state()
  client = state_obj.get_from_cache('timesketch_client')
  if not client:
    return pd.DataFrame()

  lines = []
  for line in client.get_aggregator_info():
    fields = line.get('fields', [])
    if fields:
      del line['fields']
    for index, field in enumerate(fields):
      line['field_{0:d}'.format(index+1)] = '{0:s}: {1:s}'.format(
          field.get('description', 'N/A'), field.get('name'))
    lines.append(line)
  return pd.DataFrame(lines)


# pylint:disable=unused-argument
@framework.picatrix_magic
def timesketch_get_token_status(data: Optional[Text] = '') -> Text:
  """Returns a string with information about OAUTH token expiration.

  Args:
    data (str): not used.

  Returns:
    A string with indication about OAUTH token expiration.
  """
  connect()
  state_obj = state.state()
  client = state_obj.get_from_cache('timesketch_client')
  if not client:
    return 'Not connected to a Timesketch client.'

  status = client.get_oauth_token_status()
  if status.get('expired', True):
    return 'OAUTH token has expired.'

  return 'OAUTH token has not expired, expires at: {0:s}'.format(
      status.get('expiry_time', 'Unknown Time'))


# pylint:disable=unused-argument
@framework.picatrix_magic
def timesketch_refresh_token(data: Optional[Text] = '') -> Text:
  """Refreshes OAUTH token and returns a string with information the token.

  Args:
    data (str): not used.

  Returns:
    A string with indication about OAUTH token expiration.
  """
  connect()
  state_obj = state.state()
  client = state_obj.get_from_cache('timesketch_client')
  if not client:
    return 'Not connected to a Timesketch client.'
  client.refresh_oauth_token()
  return timesketch_get_token_status('')


@framework.picatrix_magic
def timesketch_run_aggregation_dsl(
    data: Text, as_object: Optional[bool] = False) -> Union[
        pd.DataFrame, api_aggregation.Aggregation]:
  """Run an aggregation query against the datastore.

  Args:
    data (str): Elasticsearch aggregation query DSL string.
    as_object (bool): If set to True then the aggregation object
        will be returned, otherwise a pandas DataFrame (default).

  Returns:
    A pandas DataFrame with the results from the aggregation query.
  """
  connect()
  state_obj = state.state()
  sketch = state_obj.get_from_cache('timesketch_sketch')
  if not sketch:
    return 'No data, not connected to a sketch.'

  agg_obj = sketch.aggregate(data.strip())

  if as_object:
    return agg_obj

  return agg_obj.table


@framework.picatrix_magic
def timesketch_run_aggregator(
    data: Text,
    parameters: Optional[Dict[str, Any]] = None) -> api_aggregation.Aggregation:
  """Run an aggregation query against the datastore.

  Args:
    data (str): the name of the aggregator class to run.
    parameters (dict): a dict with key/value pairs of parameters
        the aggregator needs to run.

  Returns:
    Dictionary with query results or a pandas DataFrame if as_pandas
    is set to True.
  """
  connect()
  state_obj = state.state()
  sketch = state_obj.get_from_cache('timesketch_sketch')
  if not sketch:
    return 'No data, not connected to a sketch.'

  if parameters is None:
    parameters = {}

  return sketch.run_aggregator(
      data.strip(), aggregator_parameters=parameters)


# pylint: disable=unused-argument
@framework.picatrix_magic
def timesketch_get_sketches(
    data: Optional[Text] = '') -> Dict[Text, api_sketch.Sketch]:
  """Returns a dict with available sketches.

  Args:
    data (str): not used.

  Returns:
    A dict with the available sketches, with keys as sketch name and values
    as Sketch objects.
  """
  connect(ignore_sketch=True)
  state_obj = state.state()
  client = state_obj.get_from_cache('timesketch_client')

  return {x.name: x for x in client.list_sketches()}


# pylint: disable=unused-argument
@framework.picatrix_magic
def timesketch_get_searchindices(data: Optional[Text] = '') -> Dict[Text, Text]:
  """Returns a dict with available searchindices.

  Args:
    data (str): not used.

  Returns:
    A dict with the available search indices, with keys as names and values as
    search index names.
  """
  connect(ignore_sketch=True)
  state_obj = state.state()
  client = state_obj.get_from_cache('timesketch_client')

  return {x.name: x.index_name for x in client.list_searchindices()}


@framework.picatrix_magic
def timesketch_search_by_label(
    data: Text,
    return_fields: Optional[Text] = '',
    max_entries: Optional[int] = None) -> pd.DataFrame:
  """Returns a DataFrame with all events that have a certain label.

  Args:
    data (str): a string representing the label to search for.

  Returns:
      A pandas DataFrame with the search results.
  """
  return _label_search(
      data.strip(),
      return_fields=return_fields,
      max_entries=max_entries
  )


# pylint: disable=unused-argument
@framework.picatrix_magic
def timesketch_get_starred_events(
    data: Optional[Text] = '',
    return_fields: Optional[Text] = '',
    max_entries: Optional[int] = None) -> pd.DataFrame:
  """Returns a DataFrame with all starred events.

  Args:
    data (str): not used.

  Returns:
      A pandas DataFrame with starred events.
  """
  return _label_search(
      '__ts_star',
      return_fields=return_fields,
      max_entries=max_entries)


# pylint: disable=unused-argument
@framework.picatrix_magic
def timesketch_events_with_comments(
    data: Optional[Text] = '',
    return_fields: Optional[Text] = '',
    max_entries: Optional[int] = None) -> pd.DataFrame:
  """Returns a DataFrame with all events that have comments associated with it.

  Args:
    data (str): not used.

  Returns:
      A pandas DataFrame with all events that have comments associated with
      them.
  """
  return _label_search(
      '__ts_comment',
      return_fields=return_fields,
      max_entries=max_entries)


# pylint: disable=unused-argument
@framework.picatrix_magic
def timesketch_list_stories(
    data: Optional[Text] = '') -> Dict[Text, api_story.Story]:
  """Returns a dict with all the stories that are saved to the sketch.

  Args:
    data (str): not used.

  Returns:
    A dict with keys as story titles and values as Story objects.
  """
  connect(ignore_sketch=True)
  state_obj = state.state()
  sketch = state_obj.get_from_cache('timesketch_sketch')
  if not sketch:
    return 'No data, not connected to a sketch.'
  story_dict = {}
  for story in sketch.list_stories():
    story_dict[story.title] = story
  return story_dict



# -*- coding: utf-8 -*-
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the picatrix framework."""
from typing import Optional
from typing import Text

from picatrix.lib import framework
from picatrix.lib import manager


def my_very_own_test_magic(data: Text, stuff: Optional[int] = 20) -> Text:
  """This is a magic that is used for testing.

    Args:
        data (str): This is a string.
        stuff (int): And this is a number.

    Returns:
        str: A string that combines the two parameters together.
  """
  return f'{data.strip()} - {stuff}'


def test_registration():
  """Test the magic decorator."""
  magic = framework.picatrix_magic(my_very_own_test_magic)
  assert magic.__doc__.startswith('usage: %my_very_own_test_magic')
  results = magic(line='--stuff 23 this is a text')
  assert results == 'this is a text - 23'

  manager.MagicManager.deregister_magic(magic.magic_name)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for end-to-end tests."""

from typing import Callable
from typing import Generator
from typing import Text
from typing import Tuple

import unittest
import inspect
import collections
import logging

from IPython.terminal.interactiveshell import TerminalInteractiveShell


logger = logging.getLogger('picatrix.e2e_test')

# Default values based on Docker config.
TEST_DATA_DIR = '/usr/local/src/picatrix/end_to_end_tests/test_data'


class BaseEndToEndTest:
  """Base class for end to end tests.

  Attributes:
      assertions: Instance of unittest.TestCase
  """
  assertions: unittest.TestCase

  NAME = 'name'

  def __init__(self):
    """Initialize the end-to-end test object."""
    self.assertions = unittest.TestCase()
    self._counter = collections.Counter()

  def _get_test_methods(self) -> Generator[
      Tuple[Text, Callable[[TerminalInteractiveShell], None]], None, None]:
    """Inspect class and list all methods that matches the criteria.

    Yields:
        Function name and bound method.
    """
    for name, func in inspect.getmembers(self, predicate=inspect.ismethod):
      if name.startswith('test_'):
        yield name, func

  def setup(self):
    """Setup function that is run before any tests.

    This is a good place to import any data that is needed.
    """

  def run_tests(self, ip: TerminalInteractiveShell) -> collections.Counter:
    """Run all test functions from the class.

    Returns:
        Counter of number of tests and errors.
    """
    logger.info('*** %s ***', self.NAME)
    for test_name, test_func in self._get_test_methods():
      self._counter['tests'] += 1
      logger.info('Running test: %s ...', test_name)
      try:
        test_func(ip)
      except Exception:  # pylint: disable=broad-except
        logger.error(
            'Error while running test %s', self.NAME, exc_info=True)
        self._counter['errors'] += 1
        continue
      logger.info('%s [OK]', test_name)
    return self._counter



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class that defines the state, which ."""
import logging
import threading

from typing import Any
from typing import Dict
from typing import Optional
from typing import Text

from picatrix.lib import utils


logger = logging.getLogger('picatrix.state')

__state = None
_LOCK = threading.Lock()


def state(refresh_state: bool = False):
  """Property that returns a state object."""
  # pylint: disable=global-statement
  # Making sure we have only one state object.
  global __state

  with _LOCK:
    if refresh_state or __state is None:
      __state = State()

    return __state


class State:
  """Picatrix state object."""

  _cache: Dict[Text, Any] = {}
  _last_output: Any
  _last_magic: Text

  def __init__(self):
    self._last_output = None
    self._last_magic = ''

  @property
  def last_magic(self):
    """A property that returns the last magic that was run."""
    return self._last_magic

  @property
  def last_output(self):
    """A property that returns the last output from a magic."""
    return self._last_output

  def add_to_cache(self, name: Text, value: Any):
    """Add a value to the cache or update value if it exists.

    Args:
      name (str): name of the value in the cache.
      value (object): the value to be stored in the cache.
    """
    with _LOCK:
      self._cache[name] = value

  def get_from_cache(self, name: Text, default: Optional[Any] = None) -> Any:
    """Get a value from the cache.

    Args:
      name (str): name of the value in the cache to retrieve.
      default (object): if the value does not exist, defines the default
          value to return. This is optional and returns None if not defined.

    Returns:
      The value from the cache if it exists, otherwise the default value.
    """
    with _LOCK:
      return self._cache.get(name, default)

  def remove_from_cache(self, name: Text):
    """Removes a value from the cache if it exists."""
    with _LOCK:
      if name in self._cache:
        del self._cache[name]

  def set_output(
      self, output: Any, magic_name: Text,
      bind_to: Optional[Text] = '') -> Optional[Any]:
    """Sets an output from a magic and stores it in the namespace if needed.

    Args:
      output (object): the output from the magic as executed.
      magic_name (str): the name of the magic that was used.
      bind_to (str): optional name of a variable. If this is provided
        the output is omitted and variable is stored in the namespace using
        the name provided here.

    Returns:
      Returns the output object.
    """
    with _LOCK:
      self._last_output = output
      self._last_magic = magic_name

      if bind_to:
        _ = utils.ipython_bind_global(name=bind_to, value=output)
        return None

      return output



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""End to end tests of common picatrix magics."""

from IPython.terminal.interactiveshell import TerminalInteractiveShell

from timesketch_api_client import sketch

from . import interface
from . import manager


class TimesketchTest(interface.BaseEndToEndTest):
  """End to end tests for Timesketch magics functionality."""

  NAME = 'timesketch_test'

  def _setup_client(self, ip: TerminalInteractiveShell):
    """Setup the TimesketchAPI object into the IPython session."""
    ip.run_cell(raw_cell='from timesketch_api_client import client')
    res = ip.run_cell(raw_cell=(
        '_client = client.TimesketchApi(\n'
        '    host_uri="https://demo.timesketch.org",\n'
        '    username="demo",\n'
        '    password="demo",\n'
        '    verify=True,\n'
        '    auth_mode="userpass")'))
    self.assertions.assertTrue(res.success)
    res = ip.run_cell(raw_cell=(
        'from picatrix.lib import state\n'
        'state_obj = state.state()\n'
        'state_obj.add_to_cache(\'timesketch_client\', _client)\n'
    ))
    self.assertions.assertTrue(res.success)

  def _get_sketch(self, ip: TerminalInteractiveShell) -> sketch.Sketch:
    """Return a sketch object."""
    self._setup_client(ip)
    ip.run_line_magic(magic_name='timesketch_set_active_sketch', line='6')
    return ip.run_line_magic(magic_name='timesketch_get_sketch', line='')

  def test_get_sketch(self, ip: TerminalInteractiveShell):
    """Test fetching a sketch."""
    sketch_obj = self._get_sketch(ip)
    self.assertions.assertEqual(sketch_obj.id, 6)
    self.assertions.assertEqual(sketch_obj.name, 'Szechuan Sauce - Challenge')

  def test_list_views(self, ip: TerminalInteractiveShell):
    """Test listing up the available views for a sketch."""
    _ = self._get_sketch(ip)
    views = ip.run_line_magic(magic_name='timesketch_list_views', line='')
    expected_views = set([
        '18:Szechuan Hits',
        '19:Szechuan All Hits',
        '16:email_addresses'])
    self.assertions.assertEqual(set(views.keys()), expected_views)

  def test_query_data(self, ip: TerminalInteractiveShell):
    """Test querying for data in a sketch."""
    _ = self._get_sketch(ip)
    df = ip.run_line_magic(
        magic_name='timesketch_query', line=(
            '--fields datetime,origin,message,hostname,name secret AND '
            'data_type:"windows:shell_item:file_entry"'))
    df_slice = df[df.origin == 'Beth_Secret.lnk']
    self.assertions.assertTrue(df_slice.shape[0] > 0)
    origin_set = set(df.origin.unique())
    expected_set = set([
        '9b9cdc69c1c24e2b.automaticDestinations-ms', 'Beth_Secret.lnk',
        'HKEY_CURRENT_USER\\Software\\Classes\\Local Settings\\Software'
        '\\Microsoft\\Windows\\Shell\\BagMRU\\0\\0\\0',
        'NoJerry.lnk', 'PortalGunPlans.lnk', 'SECRET_beth.lnk', 'Secret.lnk',
        'Szechuan Sauce.lnk', 'f01b4d95cf55d32a.automaticDestinations-ms'])

    self.assertions.assertSetEqual(origin_set, expected_set)

  def test_context_date(self, ip: TerminalInteractiveShell):
    """Test querying for contex surrounding a date."""
    _ = self._get_sketch(ip)
    df = ip.run_line_magic(
        magic_name='timesketch_context_date', line=(
            '--minutes 10 --fields datetime,message,data_type,event_identifier'
            ',username,workstation 2020-09-18T22:24:36'))

    self.assertions.assertTrue(df.shape[0] > 5000)
    logon_df = df[df.event_identifier == 4624]
    logged_in_users = list(logon_df.username.unique())
    self.assertions.assertTrue('Administrator' in logged_in_users)
    self.assertions.assertTrue('DWM-1' in logged_in_users)

    df.sort_values('datetime', inplace=True)
    first_series = df.iloc[0]
    last_series = df.iloc[-1]

    first_time = first_series.datetime
    last_time = last_series.datetime
    delta = last_time - first_time
    delta_rounded = delta.round('min')

    # This should be 10 minutes or 600 seconds.
    self.assertions.assertTrue(delta_rounded.total_seconds() == 600.0)


manager.EndToEndTestManager.register_test(TimesketchTest)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""End to end tests of common picatrix magics."""

from IPython.terminal.interactiveshell import TerminalInteractiveShell

from . import interface
from . import manager


MAGIC_DEFINITION = (
    'from typing import Optional\n'
    'from typing import Text\n'
    '\n'
    'from picatrix.lib import framework\n'
    '\n'
    '@framework.picatrix_magic\n'
    'def my_silly_magic(data: Text, magnitude: Optional[int] = 100) -> Text:\n'
    '  """Return a silly string with no meaningful value.\n'
    '\n'
    '  Args:\n'
    '    data (str): This is a string that will be printed back.\n'
    '    magnitude (int): A number that will be displayed in the string.\n'
    '\n'
    '  Returns:\n'
    '    A string that basically combines the two options.\n'
    '  """\n'
    '  return f"This magical magic produced {magnitude} magics of '
    '{data.strip()}"\n')


class BasicTest(interface.BaseEndToEndTest):
  """End to end tests for query functionality."""

  NAME = 'basic_test'

  def test_picatrixmagics(self, ip: TerminalInteractiveShell):
    """Test the picatrixmagics."""
    magics = ip.run_line_magic(magic_name='picatrixmagics', line='')

    self.assertions.assertFalse(magics.empty)
    self.assertions.assertTrue(magics.shape[0] > 10)

  def test_magic_registration(self, ip: TerminalInteractiveShell):
    """Test registering a magic."""
    res = ip.run_cell(raw_cell=MAGIC_DEFINITION)
    self.assertions.assertTrue(res.success)

    magics = ip.run_line_magic(magic_name='picatrixmagics', line='')
    self.assertions.assertFalse(
        magics[magics.name == 'my_silly_magic'].empty)

    line = ip.run_line_magic(
        magic_name='my_silly_magic', line='--magnitude 23 this is my string')

    expected_return = (
        'This magical magic produced 23 magics of this is my string')
    self.assertions.assertEqual(line, expected_return)

manager.EndToEndTestManager.register_test(BasicTest)



# -*- coding: utf-8 -*-
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the pixatrix manager."""
import typing

import pytest
import mock

from picatrix.lib import manager
from picatrix.lib import utils


manager.get_ipython = mock.MagicMock()
utils.get_ipython = mock.MagicMock()


def test_registration():
  """Test registering a magic and getting a copy of it and de-registering."""
  manager.MagicManager.clear_magics()

  def my_magic(cell=None, line=None):
    """This is a magic."""
    if not cell:
      cell = 'foo'
    if not line:
      line = 'bar'
    return f'{cell}{line}'

  my_magic.magic_name = 'magical_function'
  my_magic.fn = my_magic
  manager.MagicManager.register_magic(my_magic)

  magic_from_manager = manager.MagicManager.get_magic('magical_function')
  assert magic_from_manager() == 'foobar'

  my_magic.magic_name = 'other_magic'
  def conditional():
    return False

  manager.MagicManager.register_magic(my_magic, conditional=conditional)
  magic_from_manager = manager.MagicManager.get_magic('other_magic')
  assert magic_from_manager is None

  manager.MagicManager.register_magic(my_magic)
  magic_from_manager = manager.MagicManager.get_magic('other_magic')
  assert magic_from_manager() == 'foobar'

  manager.MagicManager.deregister_magic('other_magic')
  magic_from_manager = manager.MagicManager.get_magic('other_magic')
  assert magic_from_manager is None

  manager.MagicManager.deregister_magic('magical_function')
  magic_from_manager = manager.MagicManager.get_magic('magical_function')
  assert magic_from_manager is None

  with pytest.raises(KeyError):
    manager.MagicManager.deregister_magic('does_not_exist')


def test_magic_info():
  """Test the get_magic_info."""
  # Start by clearing the current registration.
  manager.MagicManager.clear_magics()
  def magical_func():
    """This is a magical function that returns pure magic."""
    return 'magic'

  magical_func.magic_name = 'magical_function'
  magical_func.fn = magical_func
  manager.MagicManager.register_magic(magical_func)

  def second_magic():
    """This is even more magic."""
    return 'fab'
  second_magic.magic_name = 'some_magic'
  second_magic.fn = second_magic
  manager.MagicManager.register_magic(second_magic)

  def other_magic():
    """Could this be it?"""
    return 'true magic'

  other_magic.magic_name = 'other_magic'
  other_magic.fn = other_magic
  manager.MagicManager.register_magic(other_magic)

  info_df = manager.MagicManager.get_magic_info(as_pandas=True)
  assert len(info_df) == 3
  assert not info_df[info_df.name == 'other_magic'].empty

  desc_set = set(info_df.description.unique())
  expected_set = set([
      'Could this be it?', 'This is even more magic.',
      'This is a magical function that returns pure magic.'])

  assert desc_set == expected_set

  entries = manager.MagicManager.get_magic_info(as_pandas=False)
  assert len(entries) == 3
  names = [x[0] for  x in entries]
  assert 'other_magic' in names
  assert 'some_magic' in names


def test_helper_registration():
  """Test registering helpers."""
  def helper_func(stuff: typing.Text):
    """This is a helper function that helps."""
    back = f'{stuff}'*3
    return f'{back}\n\nwasn\'t this helpful?'

  helper_dict = typing.get_type_hints(helper_func)
  manager.MagicManager.register_helper(
      name='helper_func', helper=helper_func,
      typing_help=helper_dict)

  df = manager.MagicManager.get_helper_info()
  assert not df[df.name == 'helper_func'].empty
  series = df.set_index('name').loc['helper_func']
  assert series.arguments == 'stuff [str]'



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class that defines the manager for all magics."""

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Text
from typing import Tuple
from typing import Union

from dataclasses import dataclass
import functools
import pandas
from IPython import get_ipython

from picatrix.lib import utils
from picatrix.lib import state


@dataclass
class Helper:
  """Small structure for a helper."""
  function: Callable[..., Any]
  help: Text
  types: Dict[Text, Any]


class MagicManager:
  """Manager class for Picatrix magics."""

  MAGICS_DF_COLUMNS = ['name', 'description', 'line', 'cell', 'function']

  _magics: Dict[Text, Callable[[Text, Text], Text]] = {}
  _helpers: Dict[Text, Helper] = {}

  @classmethod
  def clear_helpers(cls):
    """Clear all helper registration."""
    for helper_name in cls._helpers:
      try:
        utils.ipython_remove_global(helper_name)
      except KeyError:
        pass
    cls._helpers = {}

  @classmethod
  def clear_magics(cls):
    """Clear all magic registration."""
    magics = list(cls._magics.keys())
    for magic_name in magics:
      cls.deregister_magic(magic_name)

  @classmethod
  def deregister_helper(cls, helper_name: Text):
    """Remove a helper from the registration.

    Args:
      helper_name (str): the name of the helper to remove.

    Raises:
      KeyError: if the helper is not registered.
    """
    if helper_name not in cls._helpers:
      raise KeyError(f'Helper [{helper_name}] is not registered.')

    _ = cls._helpers.pop(helper_name)
    try:
      utils.ipython_remove_global(helper_name)
    except KeyError:
      pass

  @classmethod
  def deregister_magic(cls, magic_name: Text):
    """Removes a magic from the registration.

    Args:
      magic_name (str): the name of the magic to remove.

    Raises:
      KeyError: if the magic is not registered.
    """
    if magic_name not in cls._magics:
      raise KeyError(f'Magic [{magic_name}] is not registered.')

    _ = cls._magics.pop(magic_name)
    try:
      utils.ipython_remove_global(f'{magic_name}_func')
    except KeyError:
      pass

    # Attempt to remove the magic definition.
    ip = get_ipython()
    magics_manager = ip.magics_manager

    if not hasattr(magics_manager, 'magics'):
      return

    line_magics = magics_manager.magics.get('line', {})
    if magic_name in line_magics:
      _ = magics_manager.magics.get('line').pop(magic_name)

    cell_magics = magics_manager.magics.get('cell', {})
    if magic_name in cell_magics:
      _ = magics_manager.magics.get('cell').pop(magic_name)

  @classmethod
  def get_helper(cls, helper_name: Text) -> Optional[Callable[..., Any]]:
    """Return a helper function from the registration."""
    return cls._magics.get(helper_name)

  @classmethod
  def get_magic(cls, magic_name: Text) -> Callable[[Text, Text], Text]:
    """Return a magic function from the registration."""
    return cls._magics.get(magic_name)

  @classmethod
  def get_helper_info(cls, as_pandas: Optional[bool] = True) -> Union[
      pandas.DataFrame, List[Tuple[Text, Text]]]:
    """Get a list of all the registered helpers.

    Args:
      as_pandas (bool): boolean to determine whether to receive the results
          as a list of tuple or a pandas DataFrame. Defaults to True.

    Returns:
        Either a pandas DataFrame or a list of tuples, depending on the
        as_pandas boolean.
    """
    if not as_pandas:
      return [(name, helper.help) for name, helper in cls._helpers.items()]

    lines = []
    for name, helper in cls._helpers.items():
      hints = helper.types
      hint_strings = []
      for key, value in hints.items():
        value_string = getattr(value, '__name__', str(value))
        hint_strings.append(f'{key} [{value_string}]')
      helper_string = ', '.join(hint_strings)

      lines.append({
          'name': name,
          'help': helper.help,
          'arguments': helper_string,
      })
    return pandas.DataFrame(lines)

  @classmethod
  def get_magic_info(cls, as_pandas: Optional[bool] = True) -> Union[
      pandas.DataFrame, List[Tuple[Text, Text]]]:
    """Get a list of all magics.

    Args:
      as_pandas (bool): boolean to determine whether to receive the results
          as a list of tuples or a pandas DataFrame. Defaults to True.

    Returns:
      Either a pandas DataFrame or a list of tuples, depending on the as_pandas
      boolean.
    """
    if not as_pandas:
      return [(x.magic_name, x.__doc__.split('\n')[0]) for x in iter(
          cls._magics.values())]

    entries = []
    for magic_name, magic_class in iter(cls._magics.items()):
      description = magic_class.__doc__.split('\n')[0]
      magic_dict = {
          'name': magic_name,
          'cell': f'%%{magic_name}',
          'line': f'%{magic_name}',
          'function': f'{magic_name}_func',
          'description': description}
      entries.append(magic_dict)

    df = pandas.DataFrame(entries)
    return df[cls.MAGICS_DF_COLUMNS].sort_values('name')

  @classmethod
  def register_helper(
      cls, name: Text, helper: Any, typing_help: Dict[Text, Any]):
    """Register a picatrix helper function.

    Args:
      name (str): the name of the helper function.
      helper (function): the helper function to register.
      typing_help (dict): dict with the arguments and their types.

    Raises:
      KeyError: if the helper is already registered.
    """
    if name in cls._helpers:
      raise KeyError(
          f'The helper [{name}] is already registered.')
    doc_string = helper.__doc__
    if doc_string:
      help_string = doc_string.split('\n')[0]
    else:
      help_string = 'No help string supplied.'

    cls._helpers[name] = Helper(
        function=helper, help=help_string, types=typing_help)

  @classmethod
  def register_magic(
      cls, function: Callable[[Text, Text], Text],
      conditional: Callable[[], bool] = None):
    """Register magic function as a magic in picatrix.

    Args:
      function (function): the function to register as a line and a
          cell magic.
      conditional (function): a function that should return a bool, used to
          determine whether to register magic or not. This can be used by
          magics to determine whether a magic should be registered or not, for
          instance basing that on whether the notebook is able to reach the
          required service, or whether a connection to a client can be achieved,
          etc. This is optional and if not provided a magic will be registered.

    Raises:
      KeyError: if the magic is already registered.
    """
    if conditional and not conditional():
      return

    magic_name = function.magic_name

    if magic_name in cls._magics:
      raise KeyError(
          f'The magic [{magic_name}] is already registered.')

    ip = get_ipython()
    if ip:
      ip.register_magic_function(
          function, magic_kind='line_cell', magic_name=magic_name)

    cls._magics[magic_name] = function
    function_name = f'{magic_name}_func'

    def capture_output(function, name):
      """A function that wraps around magic functions to capture output."""
      @functools.wraps(function)
      def wrapper(*args, **kwargs):
        function_output = function(*args, **kwargs)
        state_obj = state.state()
        return state_obj.set_output(function_output, magic_name=name)
      return wrapper

    _ = utils.ipython_bind_global(
        function_name, capture_output(function.fn, function_name))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common reusable components for picatrix."""
import sys

from typing import Any
from typing import Optional
from typing import Text

from IPython import get_ipython
from IPython.display import clear_output


def ask_question(question: Text, input_type: Optional[Any] = Text) -> Any:
  """Asks a question and returns the answer.

  Args:
    question (str): the question to be asked.
    input_type (object): the type of the input data.

  Raises:
    TypeError: if the input type is not supported.

  Returns:
    object: an object of type "input_type" read back as an answer.
  """
  print(question)
  answer_line = sys.stdin.readline()
  answer_line = answer_line.strip()

  if input_type == str:
    return answer_line

  if input_type in (int, float):
    return input_type(answer_line)

  raise TypeError('Only support str, int and float as input types')


def clear_notebook_output():
  """Clears the output cell from the notebook."""
  clear_output(wait=True)


def ipython_bind_global(name: Text, value: Any) -> Any:
  """Binds the name to a Python object denoted by value.

  Args:
    name (str): Variable name.
    value (object): Python object to bind to name

  Returns:
    Returns the value bound to the given name.
  """
  ip = get_ipython()
  if ip and name:
    ip.push({name: value})
  return value


def ipython_get_global(name: Text) -> Any:
  """Returns the Python object bound to the given name in the user namespace.

  Args:
    name (str): Variable name.

  Returns:
    Returns the value bound to the given name.

  Raises:
    KeyError: if the variable is not stored in the namespace.
  """
  ip = get_ipython()
  return ip.all_ns_refs[0][name]


def ipython_remove_global(name: Text):
  """Removes a Python object that is bound to the user namespace.

  Args:
    name (str): Variable name.

  Raises:
    KeyError: if the variable is not stored in the namespace.
  """
  ip = get_ipython()
  namespace = ip.all_ns_refs[0]

  if name not in namespace:
    raise KeyError(f'The variable {name} is not currently in the namespace.')

  _ = namespace.pop(name)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class that defines common picatrix magics."""

from typing import Any
from typing import Optional
from typing import Text

import pandas

from picatrix.lib import framework
from picatrix.lib import manager
from picatrix.lib import state


@framework.picatrix_magic
def picatrixhelpers(data: Optional[Text] = '') -> pandas.DataFrame:
  """Provides information about registered Picatrix helpers.

  Args:
    data (str): If empty an overview of all registered helpers is provided,
        otherwise the help message of a particular helper is provided.

  Returns:
      A pandas DataFrame that contains the names and basic information about
      every registered helper or information about a single helper if
      the data string is provided.
  """
  info_df = manager.MagicManager.get_helper_info(as_pandas=True)
  if not data:
    return info_df

  helper_obj = manager.MagicManager.get_helper(data)
  if not helper_obj:
    return pandas.DataFrame()

  return info_df[info_df.name == data.strip()]


@framework.picatrix_magic
def picatrixmagics(data: Optional[Text] = '') -> pandas.DataFrame:
  """Provides information about registered Picatrix magics.

  Args:
    data (str): If empty an overview of all registered magics is provided,
        otherwise the help message of a particular magic is provided.

  Returns:
      A pandas DataFrame that contains the names and basic information about
      every registered magic or information about a single magic if
      the data string is provided.
  """
  if not data:
    return manager.MagicManager.get_magic_info(as_pandas=True)

  magic_obj = manager.MagicManager.get_magic(data)

  if not magic_obj:
    return pandas.DataFrame()

  description = magic_obj.__doc__.split('\n')[0]
  return pandas.DataFrame([{
      'name': magic_obj.magic_name,
      'description': description,
      'function': '{0:s}_func'.format(magic_obj.magic_name),
      'help': magic_obj.argument_parser.format_help()}])


# pylint: disable=unused-argument
@framework.picatrix_magic
def last_output(data: Optional[Text] = '') -> Any:
  """Returns the last output from a magic that was executed.

  Args:
    data (str): optional string that does nothing.

  Returns:
    The last output from a magic that was run.
  """
  state_obj = state.state()
  output = state_obj.last_output

  return output



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class that defines the framework for exported magics and functions."""

import argparse
import functools
import inspect
import types
import typing

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Text
try:
  # Got introduced in python 3.8.
  from typing import Protocol
except ImportError:
  from typing_extensions import Protocol

from picatrix.lib import error
from picatrix.lib import manager
from picatrix.lib import state
from picatrix.lib import utils


class MagicProtocol(Protocol):
  """Simple Typing protocol class for magic functions."""

  def __call__(self, *args: Any) -> Any:
    pass


class MagicArgument:
  """Simple Argument holder for magic arguments."""

  def __init__(self, *argc, **kwargs):
    self.argc = argc
    self.kwargs = kwargs


class MagicArgumentParser(argparse.ArgumentParser):
  """Argument parser for picatrix magics."""

  def __init__(self, *args, **kwargs):
    super(MagicArgumentParser, self).__init__(*args, **kwargs)
    self.storage = {}

  def exit(self, status: Optional[int] = 0, message: Optional[Text] = ''):
    """Exiting method for argument parser.

    Args:
      status (int): exit status of the parser.
      message (str): the error message.

    Raises:
      KeyError: when the parser is unable to parse the arguments.
      error.ArgParserNonZeroStatus: when the parser has successfully completed.
    """
    if not status:
      raise error.ArgParserNonZeroStatus('Exiting.')

    if message:
      raise KeyError('Wrong usage: {0:s}'.format(message.strip()))

    raise KeyError('Wrong usage, no further error message supplied.')


class _Magic:
  """The Picatrix Magic decorator."""

  argument_parser: MagicArgumentParser
  magic_name: Text

  def __init__(
      self, fn: MagicProtocol, arguments: Optional[List[MagicArgument]] = None,
      name_func: Optional[Callable[[Text], Text]] = None):
    """Initialize the Picatrix Magic."""
    self.fn = fn
    if name_func:
      self.magic_name = name_func(fn.__name__)
    else:
      self.magic_name = fn.__name__

    functools.update_wrapper(self, fn)
    self.__name__ = self.magic_name

    # pylint: disable=access-member-before-definition
    first_line = self.__doc__.split('\n')[0]
    self.argument_parser = MagicArgumentParser(
        prog=self.magic_name,
        description=first_line,
        usage=(
            '%%%(prog)s [arguments] data\nor\n%%%%%(prog)s [arguments]\ndata'))

    if arguments:
      for argument in arguments:
        dest_name = argument.kwargs.get('dest')
        type_string = argument.kwargs.get('type')
        # We need to check whether the custom argument contains an object
        # type, if so we will need to change it into a str and store the
        # type in the argument parser so it can be confirmed after variable
        # expansion.
        if dest_name and type_string:
          if isinstance(type_string, type):
            type_string = str(type_string).split('\'')[1]
          if type_string not in ('str', 'int', 'bool', 'float', 'unicode'):
            self.argument_parser.storage[dest_name] = type_string
            argument.kwargs['type'] = str

        self.argument_parser.add_argument(*argument.argc, **argument.kwargs)
    else:
      _add_function_arguments_to_parser(fn, self.argument_parser)

    try:
      self.argument_parser.add_argument(
          '--bindto', '-bindto',
          type=str,
          default='',
          action='store',
          dest='bindto',
          help='Bind the results to a variable instead of being returned.')
    except argparse.ArgumentError:
      pass

    try:
      self.argument_parser.add_argument(
          'data', nargs='?',
          help=self.argument_parser.storage.get('_data_help', ''))
    except argparse.ArgumentError:
      pass

    self.__doc__ = self.argument_parser.format_help()

  # pylint: disable=inconsistent-return-statements
  def __call__(self, line: Text, cell: Optional[Text] = None) -> Optional[Any]:
    line_magic = cell is None

    arguments = _parse_line_string(line)

    try:
      options = self.argument_parser.parse_args(arguments)
    except KeyError as e:
      print((
          'Unable to parse arguments, with error: {0!s}.\n'
          'Correct usage is: {1:s}').format(
              e, self.argument_parser.format_help()))
      return
    except error.ArgParserNonZeroStatus:
      # When argparser ends execution but without an error this exception
      # is raised, eg: when "-h" or help is used. In those cases we need
      # to return without running the magic function.
      return

    if not line_magic:
      variable = options.data
      options.data = cell

    bind_to = options.bindto
    option_dict = options.__dict__
    del option_dict['bindto']

    # TODO: Change this so that there is no variable expansion
    # done by the core and all variable expansion is only done here.
    for key, value in iter(option_dict.items()):
      if not value:
        continue

      if not isinstance(value, str):
        continue

      if value[0] == '{' and value[-1] == '}':
        var_name = value[1:-1]
        var_type = self.argument_parser.storage.get(key)

        if '{' not in var_name and '}' not in var_name:
          var_obj = utils.ipython_get_global(var_name)
          if var_type and var_type != 'object':
            type_string = type(var_obj).__name__
            if (type_string != var_type) and not type_string.endswith(
                var_type) and not var_type.endswith(type_string):
              raise KeyError((
                  'Variable [{0:s}] is not of the correct type [{1:s}] for '
                  'this magic. Type is: {2!s}').format(
                      var_name, var_type, type(var_obj)))
          option_dict[key] = var_obj

    return_value = self.fn(**option_dict)
    state_obj = state.state()

    if not line_magic and variable:
      bind_to = variable

    return state_obj.set_output(
        output=return_value,
        magic_name=self.magic_name,
        bind_to=bind_to)

  def __dir__(self):
    options = getattr(self.argument_parser, '_option_string_actions', {})
    return list(options.keys())

  def __getattribute__(self, name: Text):
    # Overwriting function to behave like a function when called
    # by isinstance, in order to use the inspect module as well as to
    # produce a better help message.
    if name == '__class__':
      return types.FunctionType

    if name.startswith('func'):
      return self.fn.__getattribute__(name)  # pytype: disable=attribute-error

    return super(_Magic, self).__getattribute__(name)


def _get_arguments_from_arg_lines(
    arg_lines: List[Text]) -> List[Dict[Text, Text]]:
  """Return a list of parsed arguments from argument docstring.

  Args:
    arg_lines (list): a list of lines, one per argument as defined in
        the docstring.

  Returns:
    A list of dict, one per argument. Each dict will have the following
    attributes; variable, type and description.
  """
  args = []
  for arg_line in arg_lines:
    var_string, _, description = arg_line.partition(':')
    var, var_type = var_string.split()
    argument = {
        'variable': var,
        'type': var_type[1:-1],
        'description': description.strip()
    }
    args.append(argument)
  return args


def _get_argument_lines_from_docstring(lines: List[Text]) -> List[Text]:
  """Return names and types of arguments read from a doc string.

  This function takes in a function's docstring and returns back
  the argument section (Args:), with one line per argument defined.
  Arguments can be multi-lines.

  Args:
    lines (list): a list of lines read from a function's docstring.

  Returns:
    A list of lines that contain argument definitions.
  """
  arg_lines = []

  try:
    arg_index = [x.strip() for x in lines].index('Args:')
  except ValueError:
    return arg_lines

  new_index = arg_index + 1
  spaces = lines[arg_index].index('A')
  while True:
    line = lines[new_index]

    if not line.strip():
      break

    if line.strip() in ('Returns:', 'Raises:'):
      break

    # Checking indentation, whether the argument line is continuing or
    # a new argument is being defined.
    space_count = len(line) - len(line.lstrip(' '))
    if space_count == spaces * 2:
      arg_lines.append(line.strip())
    else:
      arg_lines[-1] = '{0:s} {1:s}'.format(arg_lines[-1], line.strip())

    new_index += 1
    if new_index >= len(lines)-1:
      break

  return arg_lines


def _add_function_arguments_to_parser(
    fn: MagicProtocol, parser: MagicArgumentParser):
  """Adds arguments to a parser from parsing a function.

  Args:
    fn (function): a function to extract an argument parser from.
    parser (MagicArgumentParser): an argument parser object, that will be used
        to add arguments to.

  Raises:
    ValueError: when the docstring is not correctly formatted.
  """
  try:
    doc_string_args = _parse_docstring(fn)
  except ValueError as e:
    raise ValueError(
        'Unable to register the magic %{0:s} since the docstring '
        'of the function is not correctly formatted. Error message '
        'is: {1!s}'.format(fn.__name__, e)) from e

  inspect_fn = inspect.getfullargspec

  try:
    spec = inspect_fn(fn)  # pytype: disable=wrong-arg-types
  except TypeError:
    spec = None

  if spec and spec.defaults:
    defaults_len = len(spec.defaults)
    default_values = dict(zip(spec.args[-defaults_len:], spec.defaults))
  else:
    default_values = {}

  for argument in doc_string_args:
    arg_type_string = argument.get('type', 'str')
    description = argument.get('description', '')
    # Percent sign has a special meaning for argument parser, escaping potential
    # percent signs in docstrings.
    description = description.replace('%', '%%')

    if argument.get('variable') == 'data':
      parser.storage['data'] = arg_type_string
      parser.storage['_data_help'] = description
      continue

    variable = argument.get('variable', '')
    arg_type = str
    if arg_type_string in ('str', 'unicode'):
      arg_type = str
    elif arg_type_string == 'float':
      arg_type = float
    elif arg_type_string == 'int':
      arg_type = int
    elif arg_type_string == 'bool':
      arg_type = bool
      arg_default = default_values.get(variable)
      if arg_default:
        action = 'store_false'
      else:
        action = 'store_true'
    else:
      # Type is a string reference to an object.
      arg_type = str
      parser.storage[variable] = arg_type_string

    if arg_type == bool:
      parser.add_argument(
          '--{0:s}'.format(variable), '-{0:s}'.format(variable),
          dest=variable, action=action,
          default=default_values.get(variable),
          help=description)
    else:
      parser.add_argument(
          '--{0:s}'.format(variable), '-{0:s}'.format(variable),
          dest=variable, type=arg_type, action='store',
          default=default_values.get(variable),
          help=description)


def _parse_docstring(function: MagicProtocol) -> List[Dict[Text, Text]]:
  """Return a list of arguments extracted from a function's docstring.

  Args:
    function (function): a function to extract the docstring from.

  Returns:
    A list of dict, one per argument. Each dict will have the following
    attributes; variable, type and description.
  """
  doc_string = getattr(function, '__doc__')
  if not doc_string:
    return []

  doc_lines = doc_string.split('\n')
  arg_lines = _get_argument_lines_from_docstring(doc_lines)
  return _get_arguments_from_arg_lines(arg_lines)


def _parse_line_string(line: Text) -> List[Text]:
  """Return a list of arguments from a line magic.

  Args:
    line (str): the value passed to the line magic, or line attribute in a
        cell magic.

  Returns:
    List of arguments that can be parsed by argparse.
  """
  arguments = []
  temp_args = []
  in_quotes = False
  arg_item = False
  quota_char = ''

  for item in line.strip().split():
    # Two cases to watch for, line starts with - (argument) or quotes.
    if in_quotes:
      if item[-1] == quota_char:
        temp_args.append(item[:-1])
        arguments.append(' '.join(temp_args))
        temp_args = []
        in_quotes = False
        quota_char = ''
      else:
        temp_args.append(item)
      continue

    if item[0] == '-':
      arguments.append(item)
      arg_item = True
      continue

    if item[0] in ('\'', '"'):
      quota_char = item[0]

      # The quoted argument is a single word.
      if item[-1] == quota_char:
        arguments.append(item[1:-1])
        quota_char = ''
        continue

      temp_args.append(item[1:])
      in_quotes = True
      quota_char = item[0]
      continue

    if arg_item:
      arguments.append(item)
      arg_item = False
    else:
      temp_args.append(item)

  if temp_args:
    arguments.append(' '.join(temp_args))

  return arguments


def picatrix_helper(function: Callable[..., Any]) -> Callable[..., Any]:
  """Decorator to register a picatrix helper.

  Args:
    function (function): if the decorator is called without any
        arguments the helper function is passed to the decorator.

  Returns:
   The function that was passed in.
  """
  typing_hints = typing.get_type_hints(function)
  manager.MagicManager.register_helper(
      name=function.__name__,
      helper=function,
      typing_help=typing_hints)

  try:
    _ = utils.ipython_get_global(function.__name__)
  except KeyError:
    utils.ipython_bind_global(function.__name__, function)

  return function


def picatrix_magic(
    function: Optional[MagicProtocol] = None,
    arguments: Optional[List[MagicArgument]] = None,
    name_func: Optional[Callable[[Text], Text]] = None,
    conditional: Optional[Callable[[None], bool]] = None) -> MagicProtocol:
  """Decorator to turn functions into IPYthon magics for picatrix.

  Args:
    function (function): if the decorator is called without any arguments
        the magic function is passed to the decorator.
    arguments (list): list of MagicArgument objects to pass to the magic
        argument parser.
    name_func (function: a name function that will accept a single argument
        and return back a name that will be used to register the magic.
        Optional and if not provide the name of tbe function will be used.
    conditional (function): a function that should return a bool, used to
        determine whether to register magic or not. This can be used by
        magics to determine whether a magic should be registered or not, for
        instance basing that on whether a certain magic is able to reach the
        service it requires. This is optional and if not provided a magic
        will be registered.

  Returns:
    the decorator function.
  """
  if function:
    magic_function = _Magic(
        function, name_func=name_func, arguments=arguments)
    manager.MagicManager.register_magic(magic_function, conditional)
    return magic_function

  def wrapper(func):
    """Wrapper for the magic."""
    magic_function = _Magic(
        func, name_func=name_func, arguments=arguments)

    manager.MagicManager.register_magic(magic_function, conditional)
    return magic_function

  return wrapper



# -*- coding: utf-8 -*-
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the picatrix framework."""
from typing import Optional
from typing import Text

from picatrix.lib import framework
from picatrix.lib import manager
from picatrix.magics import common


def my_very_own_test_magic(data: Text, stuff: Optional[int] = 20) -> Text:
  """This is a magic that is used for testing.

    Args:
        data (str): This is a string.
        stuff (int): And this is a number.

    Returns:
        str: A string that combines the two parameters together.
  """
  return f'{data.strip()} - {stuff}'


def test_registration():
  """Test the magic decorator."""
  manager.MagicManager.clear_magics()
  _ = framework.picatrix_magic(my_very_own_test_magic)
  magic = framework.picatrix_magic(common.picatrixmagics)

  df = magic(line='')
  assert len(df) == 2
  assert set(df.columns) == set(manager.MagicManager.MAGICS_DF_COLUMNS)
  assert set(df.name.unique()) == set(
      ['my_very_own_test_magic', 'picatrixmagics'])

  magic = framework.picatrix_magic(common.last_output)
  same_df = magic(line='')
  assert same_df.equals(df)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file contains a class for managing end to end tests."""
from typing import Generator
from typing import Tuple
from typing import Text

from . import interface


class EndToEndTestManager:
  """The test manager."""

  _class_registry: dict = {}
  _exclude_registry: set = set()

  @classmethod
  def get_tests(cls) -> Generator[
      Tuple[Text, interface.BaseEndToEndTest], None, None]:
    """Retrieves the registered tests.

    Yields:
        tuple: containing:
            str: the uniquely identifying name of the test
            type: the test class.
    """
    for test_name, test_class in iter(cls._class_registry.items()):
      if test_name in cls._exclude_registry:
        continue
      yield test_name, test_class

  @classmethod
  def get_test(cls, test_name: Text) -> interface.BaseEndToEndTest:
    """Retrieves a class object of a specific test.

    Args:
        test_name (str): name of the test to retrieve.

    Returns:
        Instance of Test class object.

    Raises:
        KeyError: if the test is not registered.
    """
    try:
      test_class = cls._class_registry[test_name.lower()]
    except KeyError as exc:
      raise KeyError(
          'No such test type: {0:s}'.format(test_name.lower())) from exc
    return test_class

  @classmethod
  def register_test(
      cls, test_class: interface.BaseEndToEndTest,
      exclude_from_list: bool = False):
    """Registers an test class.

    The test classes are identified by their lower case name.

    Args:
        test_class (type): the test class to register.
        exclude_from_list (boolean): if set to True then the test
            gets registered but will not be included in the
            get_tests function. Defaults to False.

    Raises:
        KeyError: if class is already set for the corresponding name.
    """
    test_name = test_class.NAME.lower()
    if test_name in cls._class_registry:
      raise KeyError('Class already set for name: {0:s}.'.format(
          test_class.NAME))
    cls._class_registry[test_name] = test_class
    if exclude_from_list:
      cls._exclude_registry.add(test_name)

  @classmethod
  def clear_registration(cls):
    """Clears all test registrations."""
    cls._class_registry = {}
    cls._exclude_registry = set()



# -*- coding: utf-8 -*-
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the picatrix utils."""
import io

from picatrix.lib import utils


def test_asking_questions(monkeypatch):
  """Test the utils ask_question function.."""
  expected_answer = 'picatrix'
  monkeypatch.setattr('sys.stdin', io.StringIO(expected_answer))
  text = utils.ask_question('What is your name')
  assert text == expected_answer

  expected_answer = 234.12
  monkeypatch.setattr('sys.stdin', io.StringIO('234.12'))
  number = utils.ask_question('What is your number', input_type=float)
  assert number == expected_answer



# -*- coding: utf-8 -*-
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the pixatrix state object."""
from picatrix.lib import testlib


def test_getting_last_magic():
  """Test running a magic."""
  ip = testlib.InteractiveTest.get_shell()
  ip.run_cell(raw_cell='from picatrix.magics import common')
  res = ip.run_cell(raw_cell='common.picatrixmagics("")')
  output_df = res.result
  #output_df = ip.run_line_magic(magic_name='picatrixmagics', line='')

  code = (
      'df = common.last_output("")\n'
      'names = list(df.name.unique())\n'
      'sorted(names)')

  names = list(output_df['name'].unique())
  testlib.InteractiveTest.run_and_compare(code, expected_return=sorted(names))


def test_working_with_cache():
  """Test working with the cache."""
  ip = testlib.InteractiveTest.get_shell()
  res = ip.run_cell(raw_cell=(
      'from picatrix.lib import state\n'
      'state.state()\n'))
  assert res.success
  state_obj = res.result
  state_obj.add_to_cache('foobar', 1234)

  assert state_obj.get_from_cache('foobar') == 1234

  res = ip.run_cell(raw_cell=(
      'state_obj = state.state()\n'
      'state_obj.get_from_cache("foobar")\n'))
  assert res.success
  assert res.result == 1234

  res = ip.run_cell(raw_cell='state_obj.remove_from_cache("foobar")')
  assert res.success

  res = ip.run_cell(raw_cell='state_obj.get_from_cache("foobar")')
  assert res.success
  assert res.result is None
